import java.applet.Applet;
import java.awt.BorderLayout;
import java.awt.Button;
import java.awt.Canvas;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Label;
import java.awt.MediaTracker;
import java.awt.Panel;
import java.awt.TextArea;
import java.awt.TextField;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.InputEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.util.StringTokenizer;
import java.util.Calendar;
import java.util.Vector;

import java.io.BufferedReader;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.net.URLConnection;

class Board extends Canvas
		implements MouseListener, Runnable
{
    final public int FIRST        = 1;
    final public int FAST_BACK    = 2;
    final public int BACK         = 3;
    final public int FORWARD      = 4;
    final public int FAST_FORWARD = 5;
    final public int LAST         = 6;

    final private int FAST_OFFSET = 5;

	final private String SPACE = "%20";

    /**
     * The x and y-offset for the board
     */
    private int itsXOffset = 30;
    private int itsYOffset =  5;

    /**
     * The size of the black and white stone images
     */
    private int itsSize;

    /**
     * The applet
     */
    private Applet           itsApplet;
    private Parser           itsParser;
    private Defs             itsDefs;
    private Command          itsCommand;
    private OneLineComment   itsOneLineComment;
    private MultiLineComment itsMultiLineComment;

    /**
     * The image for board, white and black stones.
     */
    private Image itsBoardImage;

    private Image itsWhiteStoneImage;
    private Image itsBlackStoneImage;

    private Image itsWhiteVariantImage;
    private Image itsBlackVariantImage;

	private Vector itsMoves;
	private int    itsMoveIndex;

	private Vector itsOriginalMoves;
	private int    itsOriginalMoveIndex;

    private Image    itsBufferedImage;
    private Graphics itsBufferedGraphics;

    private int itsImageWidth;
    private int itsImageHeight;

	private boolean isModeEdit;

    private Thread itsOnlineGameThread;
    private String itsLastOnlineGameMoves = "";

	private boolean isSearchPosition;

    public Board(Applet applet, Parser parser, Defs defs)
    {
        itsApplet = applet;
        itsParser = parser;
        itsDefs   = defs;
        itsSize   = defs.getSize();

        if (!defs.isShowCoordinates())
        {
            itsXOffset = 5;
        }
    }

    /**
     * Load images.
     */
    public void init()
    {
		try
		{
			itsBoardImage =
				itsApplet.getImage(
					itsApplet.getCodeBase(),
					"board" + itsSize + ".gif");

            itsWhiteStoneImage =
                    itsApplet.getImage(
                            itsApplet.getCodeBase(),
                            "white" + itsSize + ".gif");

            itsBlackStoneImage =
                    itsApplet.getImage(
                            itsApplet.getCodeBase(),
                            "black" + itsSize + ".gif");

            itsWhiteVariantImage =
                    itsApplet.getImage(
                            itsApplet.getCodeBase(),
                            "w.gif");

            itsBlackVariantImage =
                    itsApplet.getImage(
                            itsApplet.getCodeBase(),
                            "b.gif");
		}
		catch (Exception e)
		{
			e.printStackTrace();
		}

        MediaTracker mt = new MediaTracker(this);
        mt.addImage(itsBoardImage, 0);
        mt.addImage(itsWhiteStoneImage, 1);
        mt.addImage(itsBlackStoneImage, 2);

        try
        {
            mt.waitForAll();
        }
        catch (InterruptedException e)
        {
            e.printStackTrace();
        }

        addMouseListener(this);

		initMoves(itsParser.getMoves());

		if (itsParser.getUrlOnlineGame() != null)
		{
			start();
		}
    }

    public void initMoves(Vector moves)
    {
        itsMoves = moves;
        itsMoveIndex = itsMoves.size()-1;

		itsOriginalMoves = (Vector) itsMoves.clone();
    }

    public void setCommand(Command command)
    {
        itsCommand = command;
    }

    public void setOneLineComment(OneLineComment oneLineComment)
    {
        itsOneLineComment = oneLineComment;
    }

    public void setMultiLineComment(MultiLineComment multiLineComment)
    {
        itsMultiLineComment = multiLineComment;
    }

    /**
     * Draw the board
     * @param g The graphics handle
     */
    private void drawBoard(Graphics g)
    {
        g.setFont(itsDefs.getCoordinateFont());

        int x;
        int y;

        //
        // Draw board
        //
        if (itsDefs.isShowBoard())
        {
            g.drawImage(itsBoardImage, itsXOffset, itsYOffset, this);
        }
        else
        {
        	g.clearRect(
        		itsXOffset, itsYOffset,
        		itsBoardImage.getWidth(this), itsBoardImage.getHeight(this));
        }

        g.setColor(Color.black);

        //
        // Draw lines
        //
        for (int row = 1; row <= 15; row++)
        {
            x = itsXOffset + itsSize;
            y = itsYOffset + itsSize + itsSize * (row - 1);
            g.drawLine(x, y, x + itsSize * 14, y);
        }

        for (int column = 1; column <= 15; column++)
        {
            y = itsYOffset + itsSize;
            x = itsXOffset + itsSize + itsSize * (column - 1);
            g.drawLine(x, y, x, y + itsSize * 14);
        }

        //
        // Draw coordinates
        //
        if (itsDefs.isShowCoordinates())
        {
            for (int row = 1; row <= 15; row++)
            {
                x = itsXOffset - 20;
                y = itsYOffset + itsSize + itsSize * (15 - row) + 5;

                String number = row + "";

                if (row >= 10)
                {
                    x -= 4;
                }

                g.drawString(number, x, y);
            }

            for (int column = 1; column <= 15; column++)
            {
                String letter = Defs.getColumnImage(column);

                y = itsYOffset + 17 * itsSize - 10;
                x = itsXOffset + itsSize + itsSize * (column - 1) - 2;

                if (itsSize != Defs.BIG_SIZE)
                {
                    y += 10;
                }

                if (column >= 10)
                {
                    x -= 4;
                }

                g.drawString(letter, x, y);
            }
        }

        drawMarker(g, 8, 8);
        drawMarker(g, 4, 4);
        drawMarker(g, 4, 12);
        drawMarker(g, 12, 4);
        drawMarker(g, 12, 12);
    }

    /**
     * Draw black or white stone
     *
     * @param g The graphics handle
     * @param move The moves of the game
     */
    private void drawStone(Graphics g, Move move)
    {
        final int number = move.getNumber();
        final int column = move.getCol();
        final int row = move.getRow();

        final boolean IS_BLACK = (number & 1) != 0;

        Image stoneImage = null;
        
    	//System.out.println("column : " + column + ", row : " + row);

        if (column == 0 && row == 0)
        {
        	return; // pass move
        }

        if (IS_BLACK)
        {
            stoneImage = itsBlackStoneImage;
            g.setColor(Color.white);
        }
        else
        {
            stoneImage = itsWhiteStoneImage;
            g.setColor(Color.black);
        }

        int x = itsXOffset + column * itsSize - itsSize / 2;
        int y = itsYOffset + itsSize + itsSize * (15 - row) - itsSize / 2;

        g.drawImage(stoneImage, x, y, this);

        if (itsDefs.isShowNumbers())
        {
            String moveNumber = number + "";

            if (itsSize == Defs.BIG_SIZE)
            {
                x += 13;
                y += 20;

                if (number >= 10)
                {
                    x -= 3;
                }

                if (number >= 100)
                {
                    x -= 4;
                }
            }
            else
            {
                x += 7;
                y += 13;

                if (number >= 10)
                {
                    x -= 4;
                }

                if (number >= 100)
                {
                    x -= 3;
                }
            }

            g.setFont(itsDefs.getStoneNumberFont());
            g.drawString(moveNumber, x, y);
        }
    }

    /**
     * Draw black or white variant
     *
     * @param g The graphics handle
     * @param move The variant moves of the game
     */
    private void drawVariant(Graphics g, Variant variant, int number)
    {
		if (number - 2 != itsMoveIndex)
        {
        	return;
        }

        final int column = variant.getCol();
        final int row = variant.getRow();

        final boolean IS_BLACK = (number & 1) != 0;

        Image variantImage = null;
        
    	//System.out.println("column : " + column + ", row : " + row);

        if (IS_BLACK)
        {
            variantImage = itsBlackVariantImage;
        }
        else
        {
            variantImage = itsWhiteVariantImage;
        }

        int x = itsXOffset + column * itsSize - itsSize / 2;
        int y = itsYOffset + itsSize + itsSize * (15 - row) - itsSize / 2;

        g.drawImage(variantImage, x, y, this);
    }

    private void drawText(Graphics g, Move move)
    {
        Text[] allText = move.getText();

        if (allText != null)
        {
            for (int i = 0; i < allText.length; i++)
            {
                Text text = allText[i];

                final int column = text.getCol();
                final int row = text.getRow();
                final String str = text.getText();

                g.setColor(Color.black);

                int x = itsXOffset + column * itsSize - 5 * str.length();
                int y = itsYOffset + itsSize + itsSize * (15 - row) + 6;

                g.setFont(itsDefs.getTextFont());

                if (itsSize != Defs.BIG_SIZE)
                {
                    x += 1;
                    y -= 1;
                }

                g.drawString(str, x, y);
            }
        }
    }

    /**
     * Draw black or white stone
     *
     * @param g The graphics handle
     * @param column The column to draw the stone
     * @param row The row to draw the stone
     */
    private void drawMarker(Graphics g, int column, int row)
    {
        int x = itsXOffset + column * itsSize - 2;
        int y = itsYOffset + itsSize + itsSize * (15 - row) - 2;

        g.setColor(Color.black);
        g.fillRect(x, y, 5, 5);
    }

    /**
     * Draw one line comment
     * @param move The moves of the game
     */
    private void drawComment(Move move)
    {
    	if (itsDefs.isShowOneLineComment())
    	{
			String oneLineComment = move.getOneLineComment();

			if (oneLineComment == null)
			{
				itsOneLineComment.clear();
			}
			else
			{
				itsOneLineComment.set(oneLineComment);
			}
    	}

		if (itsDefs.isShowMultiLineComment())
		{
			String[] multiLineComment = move.getMultiLineComment();

			itsMultiLineComment.clear();

			if (multiLineComment != null)
			{
				for (int i = 0; i < multiLineComment.length; i++)
				{
					String text = multiLineComment[i];
					itsMultiLineComment.append(text);
					itsMultiLineComment.append("\n");
				}

				itsMultiLineComment.home();
			}
		}
    }

    /**
     * Paint it.
     * @param g The graphics handle
     */
    public void paint(Graphics g)
    {
        if (itsBufferedImage == null)
        {
            Dimension d = getSize();
            itsImageWidth = d.width;
            itsImageHeight = d.height;

            itsBufferedImage = createImage(itsImageWidth, itsImageHeight);
            itsBufferedGraphics = itsBufferedImage.getGraphics();
        }

        drawBoard(itsBufferedGraphics);

        for (int i = 0; i <= itsMoveIndex; i++)
        {
        	//System.out.println("paint - loop : " + i);
			Move move = (Move)itsMoves.elementAt(i);
            drawStone(itsBufferedGraphics, move);

			Variant[] variants = move.getVariant();

			if (variants != null)
			{
		        for (int v = 0; v < variants.length; v++)
		        {
		            drawVariant(
		            	itsBufferedGraphics, variants[v], move.getNumber() + 1);
		        }
			}
        }

        itsCommand.enableButtons(isFirstMove(), isLastMove());

        Move lastMove = (Move)itsMoves.elementAt(itsMoveIndex);
        drawText(itsBufferedGraphics, lastMove);

        drawComment(lastMove);

        // Display the buffered image
        g.drawImage(itsBufferedImage, 0, 0, null);

        itsCommand.showMove(lastMove);
    }

    boolean isFirstMove()
    {
        return itsMoveIndex == 0;
    }

    boolean isLastMove()
    {
        return itsMoveIndex == itsMoves.size() - 1;
    }

	public void update(Graphics g)
	{
		paint(g);
	}

	public void getOnlineMoves(boolean searchPosition)
	{
		if (isModeEdit)
		{
			return;
		}

        try
        {
        	String urlOnlineGame = itsParser.getUrlOnlineGame();

        	int index = urlOnlineGame.indexOf('?');

			if (searchPosition)
			{
				urlOnlineGame += (index == -1) ? "?" : "&";
				urlOnlineGame += "moves=" + getCurrentMoves();
			}

			System.out.println("URL : \"" + urlOnlineGame + "\"");

            URL url = new URL(urlOnlineGame);

            URLConnection connection = url.openConnection();

            InputStream is = connection.getInputStream();
            InputStreamReader isr = new InputStreamReader(is);
            BufferedReader br = new BufferedReader(isr);

            String moves = br.readLine();

            br.close();
            isr.close();

			if (moves.equals(itsLastOnlineGameMoves))
			{
				//System.out.println("--- moves.equals(itsLastOnlineGameMoves)");
				return;
			}

			//System.out.println("+++ itsLastOnlineGameMoves (new)");

			itsLastOnlineGameMoves = moves;

			itsApplet.showStatus(Calendar.getInstance().getTime().toString());

            itsParser.parseMoves(moves);
			initMoves(itsParser.getMoves());
			repaint();
        }
        catch (Exception e)
        {
            e.printStackTrace();
        }
	}

	private String getCurrentMoves()
	{
    	String moves = "";

	    for (int i = 0; i <= itsMoveIndex; i++)
	    {
			Move move = (Move)itsMoves.elementAt(i);

			moves += Defs.getCoordinateImage(move.getCol(), move.getRow());

		    if (i < itsMoveIndex)
		    {
		    	moves += SPACE;
		    }
	    }

		//System.out.println("getCurrentMoves() : \"" + moves + "\"");

		return moves;
	}

	private void updateOnlineMoves()
	{
        try
        {
        	String moves = getCurrentMoves();

            URL url = new URL(itsParser.getUrlUpdateOnlineGame() + moves);

            URLConnection connection = url.openConnection();

            InputStream is = connection.getInputStream();
            InputStreamReader isr = new InputStreamReader(is);
            BufferedReader br = new BufferedReader(isr);

            String reply = br.readLine();

            br.close();
            isr.close();
        }
        catch (Exception e)
        {
            e.printStackTrace();
        }
	}

	public void submit()
	{
		updateOnlineMoves();
	}

	public void edit()
	{
		isModeEdit = true;
		itsOriginalMoveIndex = itsMoveIndex;
	}

	public void view()
	{
		isModeEdit = false;
		itsMoveIndex = itsOriginalMoveIndex;
		itsMoves = (Vector) itsOriginalMoves.clone();
		repaint();
	}

	public boolean isVariant(int col, int row)
	{
		Move move = (Move)itsMoves.elementAt(itsMoveIndex);

		Variant[] variants = move.getVariant();

		if (variants != null)
		{
	        for (int v = 0; v < variants.length; v++)
	        {
	        	if (variants[v].getCol() == col && variants[v].getRow() == row)
	        	{
	        		return true;
	        	}
	        }
		}

		return false;
	}

	public void insertMove(int col, int row)
	{
		if (col !=0 && row !=0)
		{
			for (int i = 0; i <= itsMoveIndex; i++)
			{
				Move move = (Move)itsMoves.elementAt(i);

				if (col == move.getCol() && row == move.getRow())
				{
					return; // Not a free coordinate
				}
			}

			int nextIndex = itsMoveIndex + 1;

			if (nextIndex < itsMoves.size())
			{
				Move move = (Move)itsMoves.elementAt(nextIndex);

				if (col == move.getCol() && row == move.getRow())
				{
					move(FORWARD);
					return;
				}
			}

			for (int i = itsMoves.size()-1; i > itsMoveIndex; i--)
			{
				itsMoves.removeElementAt(i);
			}

			itsMoves.addElement(new Move(itsMoves.size()+1, col, row));
			itsMoveIndex = itsMoves.size()-1;

			repaint();
		}
	}

    public void move(int direction)
    {
        if (direction == FIRST)
        {
            itsMoveIndex = 0;
        }
        else if (direction == FAST_BACK)
        {
            itsMoveIndex = Math.max(0, itsMoveIndex - FAST_OFFSET);
        }
        else if (direction == BACK)
        {
            if (itsMoveIndex > 0)
            {
                itsMoveIndex--;
            }
        }
        else if (direction == FORWARD)
        {
            if (itsMoveIndex < itsMoves.size() - 1)
            {
                itsMoveIndex++;
            }
        }
        else if (direction == FAST_FORWARD)
        {
            itsMoveIndex = Math.min(itsMoves.size() - 1, itsMoveIndex + FAST_OFFSET);
        }
        else if (direction == LAST)
        {
            itsMoveIndex = itsMoves.size() - 1;
        }

        repaint();
    }

    /**
     * The user has clicked in the applet.
     * @param e The mouse event
     */
    public void mouseReleased(MouseEvent e)
    {
    }

    /**
     * The user has clicked in the applet.
     * @param e The mouse event
     */
    public void mousePressed(MouseEvent e)
    {
        if (itsDefs.isShowCommands())
        {
            int modifiers = e.getModifiers();

			int col = 0;
			int row = 0;

			int x = e.getX() - itsXOffset - itsSize/2;
			int y = e.getY() - itsYOffset/2 - itsSize/2;

			if ((x > 0) && (x < itsSize * 15) && (y > 0) && (y < itsSize * 15))
			{
				col = x/itsSize + 1;
				row = 15 - (y/itsSize);
			}

            if ((modifiers & InputEvent.BUTTON1_MASK) != 0)
            {
                // Left button

                if (isModeEdit)
                {
                    insertMove(col, row);
                }
                else if (isVariant(col, row))
                {
                    insertMove(col, row);
                    isSearchPosition = true;
                    getOnlineMoves(isSearchPosition);
                }
                else
                {
                    move(FORWARD);
                }
            }
            else if ((modifiers & InputEvent.BUTTON3_MASK) != 0)
            {
                // Right button
                move(BACK);
            }
        }
    }

    public void mouseClicked(MouseEvent e)
    {
    }

    public void mouseEntered(MouseEvent e)
    {
    }

    public void mouseExited(MouseEvent e)
    {
    }

    private void start()
    {
        itsOnlineGameThread = new Thread(this);
        itsOnlineGameThread.start();
    }

	public void stop()
	{
		try
		{
			itsOnlineGameThread = null;
		}
		catch (Exception e)
		{
			// Null pointer exception in case applet is destroyed
			// due to change of web page
		}
	}

    public void run()
    {
    	//System.out.println("Thread started: " + Calendar.getInstance().getTime());

        while (itsOnlineGameThread != null)
        {
            try
            {
            	//System.out.println("Thread time: " + Calendar.getInstance().getTime());
				getOnlineMoves(isSearchPosition);

				//System.out.println("Thread.sleep(" + itsDefs.getRefreshInterval() + ")");
                Thread.sleep(itsDefs.getRefreshInterval());
            }
            catch (InterruptedException e)
            {
                break;
            }
        }

        stop();
    }

    public void destroy()
    {
        removeMouseListener(this);

		if (itsOnlineGameThread != null)
		{
			itsOnlineGameThread.interrupt();
		}
	}
}
